/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#ifndef POSIX_ENDIAN_INCLUDED
#define POSIX_ENDIAN_INCLUDED

#include <cinttypes>
#include <endian.h>

#include <cassert>

namespace posix
{

inline uint32_t htobef(float value)
{
	union
	{
		float real;
		uint32_t uint;
	}
	conv;
	conv.real = value;

#if __cplusplus < 201103L
	return htobe32(conv.uint);
#else
	return htobe32(conv.uint);
#endif
}

inline float betohf(uint32_t value)
{
	union
	{
		float real;
		uint32_t uint;
	}
	conv;
	conv.uint = be32toh(value);
	return conv.real;
}

/**
 * @brief static endianness converter
 */
class Endian
{
public:

	union real32_int32
	{
		float real;
		uint32_t integer;
	};

	union real64_int64
	{
		double real;
		uint64_t integer;
	};

	static uint8_t htobe(uint8_t value)
	{
		return value;
	}

	static int8_t htobe(int8_t value)
	{
		return value;
	}

	static uint16_t htobe(uint16_t value)
	{
		return htobe16(value);
	}

	static int16_t htobe(int16_t value)
	{
		return htobe16(value);
	}

	static uint32_t htobe(uint32_t value)
	{
		return htobe32(value);
	}

	static int32_t htobe(int32_t value)
	{
		return htobe32(value);
	}

	static uint64_t htobe(uint64_t value)
	{
		return htobe64(value);
	}

	static int64_t htobe(int64_t value)
	{
		return htobe64(value);
	}

	static int32_t htobe(float value)
	{
		assert(sizeof (uint32_t) == sizeof (float));

#if __cplusplus < 201103L
		real32_int32 buf;
		buf.real = value;
#else
		real32_int32 buf = {.real = value};
#endif

		buf.integer = htobe(buf.integer);
		return buf.integer;
	}

	static int64_t htobe(double value)
	{
		assert(sizeof (uint64_t) == sizeof (double));

#if __cplusplus < 201103L
		real64_int64 buf;
		buf.real = value;
#else
		real64_int64 buf = {.real = value};
#endif

		buf.integer = htobe(buf.integer);
		return buf.integer;
	}

	static uint8_t betoh(uint8_t value)
	{
		return value;
	}

	static int8_t betoh(int8_t value)
	{
		return value;
	}

	static uint16_t betoh(uint16_t value)
	{
		return be16toh(value);
	}

	static int16_t betoh(int16_t value)
	{
		return be16toh(value);
	}

	static uint32_t betoh(uint32_t value)
	{
		return be32toh(value);
	}

	static int32_t betoh(int32_t value)
	{
		return be32toh(value);
	}

	static uint64_t betoh(uint64_t value)
	{
		return be64toh(value);
	}

	static int64_t betoh(int64_t value)
	{
		return be64toh(value);
	}

	static float betoh(float value)
	{
		assert(sizeof (uint32_t) == sizeof (float));

#if __cplusplus < 201103L
		real32_int32 buf;
		buf.real = value;
#else
		real32_int32 buf = {.real = value};
#endif

		buf.integer = be32toh(buf.integer);
		return buf.real;
	}

	static double betoh(double value)
	{
		assert(sizeof (uint64_t) == sizeof (double));

#if __cplusplus < 201103L
		real64_int64 buf;
		buf.real = value;
#else
		real64_int64 buf = {.real = value};
#endif
		buf.integer = be64toh(buf.integer);
		return buf.real;
	}
};

}

#endif // POSIX_ENDIAN
