/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#ifndef POSIX_IO_HPP_INCLUDED
#define POSIX_IO_HPP_INCLUDED

#include <unistd.h>
#include <fcntl.h>
#include <poll.h>

#include <cstdlib>

#include <vector>

#include "posixcpp.hpp"
#include "posixcpp_bytearray.hpp"
#include "posixcpp_flags.hpp"
#include "posix_time_timespec.hpp"

POSIX_BEGIN_NAMESPACE

/**
 * @brief Package for objects, envolved in the I/O process
 */
namespace io
{
#if __cplusplus >= 201103L
/**
 * @brief Read/Write mode
 */
POSIXCPP_ENUM(IOMode_t)
{
	/**
	 * I/O operation returns like C single read/write, with the report of
	 * actually received/transmitted bytes
	 */
	SINGLE,
	/**
	 * I/O operation make cycles, trying to receive/transmit requested data ammount
	 */
	LOOPED
};
/**
 * @brief Node access mode
 */
POSIXCPP_ENUM(AccessMode_t)
{
	READ_ONLY = O_RDONLY,
	READ_WRITE = O_RDWR,
	WRITE_ONLY = O_WRONLY
};

POSIXCPP_FLAGS(CreateFlags, CreateMode_t, int)
{
	NONE =			0,
	CREAT =			O_CREAT,
	NOCTTY =		O_NOCTTY,
	CLOSE_ON_EXEC =		O_CLOEXEC,
	DIRECT =		O_DIRECT,
	NO_ACCESS_TIME =	O_NOATIME,
	DIRECTORY =		O_DIRECTORY,
	EXCL =			O_EXCL
};

/**
 * @brief Events for blocking file operations
 */
POSIXCPP_FLAGS(Events, Events_t, short)
{
	/**
	 * Data ready to read
	 */
	INPUT = POLLIN,
	/**
	 * Data ready to write
	 */
	OUTPUT = POLLOUT,
	/**
	 * High priority data ready
	 */
	URGENT = POLLPRI,
	/**
	 * Error occured
	 */
	ERROR = POLLERR,
	HANGUP = POLLHUP,
	/**
	 * Invalid request
	 */
	INVALID = POLLNVAL
};

/**
 * @brief Access mode for file or node
 */
POSIXCPP_FLAGS(Mode, Mode_t, mode_t)
{
	NONE = 0,
	RWXU = S_IRWXU,
	RUSR = S_IRUSR,
	WUSR = S_IRUSR,
	XUSR = S_IXUSR,
	RWXG = S_IRWXG,
	RGRP = S_IRGRP,
	WGRP = S_IWGRP,
	XGRP = S_IXGRP,
	RWXO = S_IRWXO,
	ROTH = S_IROTH,
	WOTH = S_IWOTH,
	XOTH = S_IXOTH
};

/**
 * @brief blocking mode for opening the file
 */
POSIXCPP_ENUM(BlockingMode_t)
{
	BLOCK = 0,
	NON_BLOCK = O_NONBLOCK
};
/**
 * @brief ASCII control characters codes
 */
POSIXCPP_SIZED_ENUM(ControlCharacters_t, uint8_t)
{
	/**
	 * NULL
	 */
	NUL = 0x00,
	/**
	 * Start of header
	 */
	SOH = 0x01,
	/**
	 * Start of text
	 */
	STX = 0x02,
	/**
	 * End of text
	 */
	ETX = 0x03,
	/**
	 * End of transmission
	 */
	EOT = 0x04,
	/**
	 * End of query
	 */
	ENQ = 0x05,
	/**
	 * Acknowledge
	 */
	ACK = 0x06,
	/**
	 * Bell
	 */
	BEL = 0x07,
	/**
	 * Backspace
	 */
	BS = 0x08,
	/**
	 * Horizontal tab
	 */
	HT = 0x09,
	/**
	 * Line feed
	 */
	LF = 0x0A,
	/**
	 * Vertical tab
	 */
	VT = 0x0B,
	/**
	 * Form feed
	 */
	FF = 0x0C,
	/**
	 * Carriage return
	 */
	CR = 0x0D,
	/**
	 * Not acknowledge
	 */
	NAK = 0x15,
	/**
	 * Cancel transmission
	 */
	CAN = 0x18
};
#else
struct IOMode_t
{
enum Enum
{
	/**
	 * I/O operation returns like C single read/write, with the report of
	 * actually received/transmitted bytes
	 */
	SINGLE,
	/**
	 * I/O operation make cycles, trying to receive/transmit requested data ammount
	 */
	LOOPED
};
};

struct AccessMode{
	enum AccessMode_t
	{
		READ_ONLY = O_RDONLY,
		READ_WRITE = O_RDWR,
		WRITE_ONLY = O_WRONLY,
	};
};

struct Events
{
	enum Events_t
	{
		/**
		 * Data ready to read
		 */
		INPUT = POLLIN,
		/**
		 * Data ready to write
		 */
		OUTPUT = POLLOUT,
		/**
		 * High priority data ready
		 */
		URGENT = POLLPRI,
		/**
		 * Error occured
		 */
		ERROR = POLLERR,
		HANGUP = POLLHUP,
		/**
		 * Invalid request
		 */
		INVALID = POLLNVAL
	};
};

enum Mode_t
{
	NONE = 0,
	RWXU = S_IRWXU,
	RUSR = S_IRUSR,
	WUSR = S_IRUSR,
	XUSR = S_IXUSR,
	RWXG = S_IRWXG,
	RGRP = S_IRGRP,
	WGRP = S_IWGRP,
	XGRP = S_IXGRP,
	RWXO = S_IRWXO,
	ROTH = S_IROTH,
	WOTH = S_IWOTH,
	XOTH = S_IXOTH
};

enum BlockingMode_t
{
	BLOCK = 0,
	NON_BLOCK = O_NONBLOCK
};

/**
 * @brief ASCII control characters codes
 */
struct ControlCharacters_t
{
	enum Enum
	{
		/**
		 * NULL
		 */
		NUL = 0x00,
		/**
		 * Start of header
		 */
		SOH = 0x01,
		/**
		 * Start of text
		 */
		STX = 0x02,
		/**
		 * End of text
		 */
		ETX = 0x03,
		/**
		 * End of transmission
		 */
		EOT = 0x04,
		/**
		 * End of query
		 */
		ENQ = 0x05,
		/**
		 * Acknowledge
		 */
		ACK = 0x06,
		/**
		 * Bell
		 */
		BEL = 0x07,
		/**
		 * Backspace
		 */
		BS = 0x08,
		/**
		 * Horizontal tab
		 */
		HT = 0x09,
		/**
		 * Line feed
		 */
		LF = 0x0A,
		/**
		 * Vertical tab
		 */
		VT = 0x0B,
		/**
		 * Form feed
		 */
		FF = 0x0C,
		/**
		 * Carriage return
		 */
		CR = 0x0D,
		/**
		 * Not acknowledge
		 */
		NAK = 0x15,
		/**
		 * Cancel transmission
		 */
		CAN = 0x18
	};
};
#endif

/**
 * @brief Data buffer for scatter/gather IO
 */
class Vector
{
public:
	
	Vector(const posixcpp::ByteArray &data) :
	_data {const_cast<posixcpp::ByteArray&>(data).data(), data.size()}
	{
	}
	
	Vector(void *base, size_t len) :
	_data {base, len}
	{
	}
	
	Vector &operator =(const posixcpp::ByteArray &data)
	{
		_data.iov_base = const_cast<posixcpp::ByteArray&>(data).data();
		_data.iov_len = data.size();
		
		return (*this);
	}
	
	iovec data() const { return (_data); }
	iovec &data() { return (_data); }
private:
	iovec	_data;
};

static_assert (sizeof (Vector) == sizeof (iovec), "bad size");

/**
 * @brief File operations interface
 */
class IHandle
{
	POSIXCPP_INTERFACE(IHandle)
public:
	
	/**
	 * @brief Close file
	 */
	virtual void close() = 0;
	
	virtual void read(posixcpp::ByteArray&) = 0;
	
	/**
	 * @brief Read data from file
	 * @param data Byte array to read to, can be reduced in size to actual read count
	 * @param mode Read mode ( single system call / looped untill the glorious end )
	 * @param timeout poll timeout for looped mode
	 */
	virtual void read(
	    posixcpp::ByteArray & data,
#if __cplusplus >= 201103L
	    IOMode_t mode,
#else
	    IOMode_t::Enum mode = IOMode_t::SINGLE,
#endif
	    time::Timespec timeout = -1) = 0;

	/**
	 * @brief Scatter read
	 * @param vec buffers array
	 */
	virtual ssize_t read(const std::vector<Vector>&) = 0;
	
	virtual void read(const std::vector<posixcpp::ByteArray*>&) = 0;
	
	/**
	 * @brief Simple write
	 * @param data buffer to write
	 * @return bytes written
	 */
	virtual ssize_t write(const posixcpp::ByteArray&) = 0;
	
	virtual ssize_t write(const std::vector<const posixcpp::ByteArray*>&) = 0;
	
	/**
	 * @brief gather write
	 * @param vec vector of buffers to write in one syscall
	 */
	virtual ssize_t write(const std::vector<Vector>&) = 0;
	
	/**
	 * @brief Write data to file
	 * @param data Byte array to write from
	 * @param mode Write mode ( single system call / looped untill the glorious end )
	 * @param timeout poll timeout for looped mode
	 * @return actual read bytes
	 */
	virtual ssize_t write(
	    const posixcpp::ByteArray&,
#if __cplusplus >= 201103L
	    IOMode_t mode,
#else
	    IOMode_t::Enum mode = IOMode_t::LOOPED,
#endif
	    time::Timespec timeout = -1) = 0;

	/**
	 * @brief Poll the device
	 * @param sevents Events wait to
	 * @param revents occured events
	 * @param timeout
	 * @return true if any events occured, false on timeout
	 */
	virtual bool poll(
	    Events sevents,
	    Events &revents,
	    time::Timespec timeout = -1) = 0;

	virtual BlockingMode_t blockingMode() const = 0;

	virtual void setBlockingMode(BlockingMode_t newVal) = 0;
};

class Handle;
/**
 * @brief make the anonymous channel (pipe)
 * 
 * @param OUT: first first pipe end
 * @param OUT: second second pipe end
 */
void pipe(Handle &first, Handle &second);

class Pipe;

class IFile;

class File;

class FIFO;

class TTY;

class PseudoTTY;

class I2C;
	
/**
 * Default numer of receive loop iterations to receive requested data size
 */
const unsigned DEFAULT_RECEIVETRIES = 15;
/**
 * Default numer of send loop iterations to send requested data size
 */
const unsigned DEFAULT_SENDTRIES = 15;

std::ostream &operator<<(
    std::ostream &stream,
    ControlCharacters_t cChar);

}

POSIX_END_NAMESPACE

#endif // POSIX_IO_HPP_INCLUDED
