/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#ifndef POSIX_IO_HANDLE_HPP_INCLUDED
#define POSIX_IO_HANDLE_HPP_INCLUDED

#include "posix_io.hpp"
#include <vector>

POSIX_BEGIN_NAMESPACE

namespace io
{
/**
 * @brief Abstract class, implements common operations for ordinar files, fifos,
 * ttys, pseudottys and sockets
 */
class Handle : public virtual IHandle
{
	
	friend void pipe(Handle&, Handle&);
public:
	
	/**
	 * @brief Default c-tor, makes the not functional handle with -1 value
	 */
	Handle();
	
	/**
	 * @brief C-tor, wrapping the object around the exicting file number
	 * @param IN: descriptor existing file/socket descriptor number
	 */
	Handle(int);
	
	/**
	 * @brief Initializeing c-tor
	 */
	Handle(BlockingMode_t bMode, size_t readTries, size_t writeTries);
	
	virtual ~Handle() noexcept = default;
	
	/**
	 * @brief Close the handle and set the number to -1
	 */
	void close();
	
	void read(posixcpp::ByteArray&) override;
	
	/**
	 * @brief Read the handle to bytearray object
	 * 
	 * The buffer size defines the maximum bytes number to read.
	 * As a result, buffer size is set to actual read bytes count.
	 * 
	 * @param data IN, OUT: ByteArray buffer to read.
	 * @param iomode IN: Sets the mode (simple read or using the looped helper)
	 * @param timeout IN: Timeout for looped mode cycles. Default value sets
	 *     the endleess timeout
	 */
	void read(posixcpp::ByteArray&, IOMode_t iomode,
	    time::Timespec timeout = -1000000) override;

	ssize_t read(const std::vector<Vector>&) override;
	
	void read(const std::vector<posixcpp::ByteArray*>&) override;
	
	/**
	 * @brief overrides
	 */
	ssize_t write(const posixcpp::ByteArray&) override;

	/**
	 * @brief Write data to the handle object
	 * 
	 * @param data IN: ByteArray buffer to read.
	 * @param iomode IN: Sets the mode (simple write or using the looped helper)
	 * @param timeout IN: Timeout for looped mode cycles. Default value sets
	 *     the endleess timeout
	 */
	ssize_t write(const posixcpp::ByteArray&, IOMode_t mode,
	    time::Timespec timeout = -1000000) override;
	
	ssize_t write(const std::vector<Vector>&) override;
	
	ssize_t write(const std::vector<const posixcpp::ByteArray*>&) override;

	bool poll(Events sevents, Events &revents, time::Timespec timeout = -1) override;

	static std::pair<Handle, Handle> pipe();

	/**
	 * @property Descriptor
	 * @return POSIX file/socket descriptor
	 */

	int descriptor() const
	{
		return (_descriptor);
	}

	void setDescriptor(int newVal)
	{
		_descriptor = newVal;
	}

	/**
	 * @property ReadTries
	 * @brief Maximal number of receive iterations for looped mode
	 */

	unsigned readTries() const
	{
		return (_readTries);
	}

	void setReadTries(unsigned newVal)
	{
		_readTries = newVal;
	}

	/**
	 * @property WriteTries
	 * @brief Maximal number of write iterations for looped mode
	 */

	size_t writeTries() const
	{
		return (_writeTries);
	}

	void setWriteTries(size_t newVal)
	{
		_writeTries = newVal;
	}

	/**
	 * @property blockingMode
	 */

	BlockingMode_t blockingMode() const
	{
		return (_blockingMode);
	}

	void setBlockingMode(BlockingMode_t newVal)
	{
		_blockingMode = newVal;
	}

private:
	/**
	 * File/Socket descriptor
	 */
	int		_descriptor;

	/**
	 * Number of the looped mode receive tries
	 */
	size_t		_readTries;

	/**
	 * Number of the looped mode send tries
	 */
	size_t		_writeTries;

	/**
	 * Blocking mode
	 */
	BlockingMode_t	_blockingMode;
};

/**
 * @brief anonymous channel
 */
class Pipe
{
	POSIXCPP_NOT_COPYABLE(Pipe)
public:
	Pipe();
	Pipe(AccessMode_t, BlockingMode_t bMode = BlockingMode_t::BLOCK);
	
	Handle &readPoint()
	{
		return (_handle[0]);
	}
	
	Handle &writePoint()
	{
		return (_handle[1]);
	}
private:
	int	_pipeFds[2];
	Handle	_handle[2];
};

}

POSIX_END_NAMESPACE

#endif // POSIX_POSIX_IO_HANDLE_HPP_INCLUDED
