/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#ifndef POSIX_SIGNALS_HPP
#define	POSIX_SIGNALS_HPP

#include <csignal>
#include <ostream>
#include <list>

namespace posix
{
namespace signals
{

enum Number_t
{
	UNKNOWN = 0,
	HUP = SIGHUP,
	INT = SIGINT,
	QUIT = SIGQUIT,
	ILL = SIGILL,
	TRAP = SIGTRAP,
	ABRT = SIGABRT,
	BUS = SIGBUS,
	FPE = SIGFPE,
	KILL = SIGKILL,
	SEGV = SIGSEGV,
	USR1 = SIGUSR1,
	PIPE = SIGPIPE,
	USR2 = SIGUSR2,
	TERM = SIGTERM,
	SIGNALS_NUMBER
};

class Signal
{
public:
    
	Signal();
    
	Signal(Number_t number);
    
	Signal(const siginfo_t & siginfo);
    
	Number_t	 sigNumber() const;
	
	int		 errNumber() const;
	
	void		*pointer() const;
    
	void setSignumber(Number_t value);
    
	siginfo_t & siginfo()
	{
		return (m_sigInfo);
	}
    
	const siginfo_t & siginfo() const
	{
		return (m_sigInfo);
	}
    
protected:
	
	friend std::ostream & operator <<(
	  std::ostream &stream,
	  const Signal &entity);
	
private:
	
	siginfo_t m_sigInfo;
};

class Handler
{
public:
	
	/**
	 * Type of the signal handler
	 */
	enum HandleType
	{
		/**
		 * Simple handler, runs in the signal callback
		 */
		BASIC,
		/**
		 * Queued handler, runs in the application workcycle
		 */
		QUEUED
	};
	
	Handler();
	
	~Handler();
	
	void switchSignalOn(Number_t, HandleType);
	
	void sigQueue(Number_t);
	
protected:
	
	virtual void	signalAction(const Signal&) = 0;
	
private:
	
	struct HandleInfo
	{
		HandleType	type;
		
		void		(Handler::*handler)(const Signal&);
	};
	
	typedef		std::list<Signal> Signals_t;
	
	static void	_action(int, siginfo_t*, void*);
	
	void		action(const Signal&);
	
	Signals_t	_signalsQueue;
	
	HandleInfo	*_handlersInfo;
};

std::ostream & operator <<(
  std::ostream &stream,
  Number_t data);
}

}

#endif	/* POSIX_SIGNALS_HPP */
