/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#ifndef POSIX_THREADS_HPP_INCLUDED
#define POSIX_THREADS_HPP_INCLUDED

#include <pthread.h>

#include "posixcpp.hpp"

namespace posix
{

/**
 * @brief Namespace of threading functions
 */
namespace threads
{

/**
 * @enum SchedulerPolicy_t
 * @brief Thread scheduling policy enumeration
 */
POSIXCPP_ENUM(SchedulerPolicy_t)
{
	/**
	* Default OS specific policy
	*/
	DEFAULT = SCHED_OTHER,
	/**
	* Quant based real-time policy
	*/
	ROUND_ROBIN = SCHED_RR,
	/**
	* First in first out real-time policy
	*/
	FIFO = SCHED_FIFO
};

/**
 * @struct SchedulerParameters_t
 * @brief parameters of the threads scheduler
 */
struct SchedulerParameters_t
{
	SchedulerPolicy_t policy;
	int priority;
};

/**
 * @enum CancelState_t
 * @brief Cancel state of the thread
 */
POSIXCPP_ENUM(CancelState_t)
{
	/**
	* Cancelation enabled
	*/
	ENABLE = PTHREAD_CANCEL_ENABLE,
	/**
	* Cancelation disabled
	*/
	DISABLE = PTHREAD_CANCEL_DISABLE
};

/**
 * @brief Process shared or private flag.
 */
POSIXCPP_ENUM(ProcessShared_t)
{
	SHARED = PTHREAD_PROCESS_SHARED,
	PRIVATE = PTHREAD_PROCESS_PRIVATE
};

/**
 * @enum CancelType_t
 * @brief Cancelation request processing type
 */
POSIXCPP_ENUM(CancelType_t)
{
	/**
	 * The cancelation request should be deferred until the cancelation-point function
	 * will be executed
	 */
	DEFERRED = PTHREAD_CANCEL_DEFERRED,
	/**
	 * Dangerous asynchronous cancelation
	 */
	ASYNCHRONOUS = PTHREAD_CANCEL_ASYNCHRONOUS
};

/**
 * @enum SchedulerInheritance_t
 * @brief Scheduler inheritance mode for the spawning threads
 */
POSIXCPP_ENUM(SchedulerInheritance_t)
{
	INHERIT = PTHREAD_INHERIT_SCHED,
	EXPLICIT = PTHREAD_EXPLICIT_SCHED
};
/*
class IInitializer
{
	POSIXCPP_INTERFACE(IInitializer)
public:
	virtual void initialize() = 0;
};

class Once
{
public:
	Once();
	Once(IInitializer&);
	virtual ~Once();
	
	const pthread_once_t	&handle() const
	{
		return (_handle);
	}
	pthread_once_t	&handle()
	{
		return (_handle);
	}
protected:
	virtual void initialize();
private:
	static void _initialize();
	pthread_once_t	_handle;
};
*/
std::ostream &operator<<(std::ostream&, SchedulerPolicy_t);

std::ostream &operator<<(std::ostream&, SchedulerParameters_t);

std::ostream &operator<<(std::ostream&, CancelState_t);

std::ostream &operator<<(std::ostream&, CancelType_t);

std::ostream &operator<<(std::ostream&, SchedulerInheritance_t);
}
}

#endif // POSIX_THREADS_HPP_INCLUDED
