/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#include "posix_io_handle.hpp"

#include <cassert>
#include <sys/uio.h>
#include <vector>

using namespace posix::io;
using namespace posix;

Handle::Handle() :
_descriptor(-1),
_readTries(DEFAULT_RECEIVETRIES),
_writeTries(DEFAULT_SENDTRIES),
_blockingMode(BlockingMode_t::NON_BLOCK)
{
}

Handle::Handle(int descr) :
_descriptor(descr),
_readTries(DEFAULT_RECEIVETRIES),
_writeTries(DEFAULT_SENDTRIES),
_blockingMode(BlockingMode_t::NON_BLOCK)
{
}

Handle::Handle(BlockingMode_t bMode, size_t readTries, size_t writeTries) :
_descriptor(-1),
_readTries(readTries),
_writeTries(writeTries),
_blockingMode(bMode)
{
}

void
Handle::close()
{
	PRECONDITION_ASSERT(descriptor() > -1, "node is invalid");

	if (_descriptor >= 0) {
		POSIX_CALL_ASSERT_X(::close(_descriptor) != -1,
		    "Unable to close the node");
		_descriptor = -1;
	}
}

void
Handle::read(const std::vector<posixcpp::ByteArray*> &data)
{
	std::vector<Vector> v;
	v.reserve(data.size());
	for (posixcpp::ByteArray *ba : data)
		v.emplace_back(*ba);
	
	ssize_t count = ::readv(this->descriptor(),
	    reinterpret_cast<iovec*>(v.data()), v.size());
	if (count > 0) {
		for (posixcpp::ByteArray *ba : data) {
			if (count < ssize_t(ba->size())) {
				ba->resize(count);
				count = 0;
			} else
				count -= ba->size();
		}
	} else {
		for (posixcpp::ByteArray *ba : data)
			ba->resize(0);
		
		if ((count < 0) && (errno != EAGAIN) &&
		    (errno != EWOULDBLOCK))
			// Error while reading
			throw std::system_error(errno,
			    std::system_category(),
			    "Unable to read specified data");
	}
}

void
Handle::read(posixcpp::ByteArray &data)
{
	PRECONDITION_ASSERT(!data.isEmpty(), "data is empty");
	PRECONDITION_ASSERT(descriptor() > -1, "file isn't open");

	// Number of bytes have been read
	ssize_t count;

	count = ::read(_descriptor, data.data(), data.size());
	// Read was success
	if (count > 0) {
		POSTCONDITION_ASSERT(size_t(count) <= data.size(),
		    "Too much data was read");
		// Read required data size
		if (size_t(count) == data.size())
			return;
		//Read less then required
		else if (size_t(count) < data.size())
			data.resize(count);
	} else {
		data.resize(0);
		if ((count < 0) && (errno != EAGAIN) &&
		    (errno != EWOULDBLOCK))
			// Error while reading
			throw std::system_error(errno,
			    std::system_category(),
			    "Unable to read specified data");
	}
}

ssize_t
Handle::read(const std::vector<Vector> &data)
{
	return (::readv(this->descriptor(), reinterpret_cast<const iovec*>(data.data()),
	    data.size()));
}

void
Handle::read(posixcpp::ByteArray &data, IOMode_t mode,
    time::Timespec timeout)
{
	PRECONDITION_ASSERT(!data.isEmpty(), "data is empty");
	PRECONDITION_ASSERT(descriptor() > -1, "file isn't open");

	// Number of bytes have been read
	ssize_t count;

	// Single read mode is set
	if (mode == IOMode_t::SINGLE) {
		count = ::read(_descriptor, data.data(), data.size());
		// Read was success
		if (count > 0) {
			POSTCONDITION_ASSERT(size_t(count) <= data.size(),
			    "Too much data was read");
			// Read required data size
			if (size_t(count) == data.size())
				return;
				//Read less then required
			else if (size_t(count) < data.size())
				data.resize(count);
		} else {
			data.resize(0);
			if ((count < 0) && (errno != EAGAIN) &&
			    (errno != EWOULDBLOCK))
				// Error while reading
				throw std::system_error(errno,
				    std::system_category(),
				    "Unable to read specified data");
		}
	} else {
		// Cyclic read mode is set
		//Remaining read tries
		ssize_t receiveTries = _readTries;
		// Read data size
		size_t total = 0UL;
		// I/O events
		Events src, res;

		bool evFlag;
		do {
			do {
				src.clear();
				res.clear();
				src.setFlag(Events_t::INPUT);
				evFlag = this->poll(src, res, timeout);
			} while (evFlag && !res.testFlag(Events_t::INPUT));
			count = ::read(_descriptor, data.data() + total,
			    data.size() - total);
			// If success reqading
			if (count > 0) {
				total += count;
				POSTCONDITION_ASSERT(total <= data.size(),
				    "Too much data was read");
			} else if ((errno != EAGAIN) && (errno != EWOULDBLOCK)
			    && (count < 0)) {
				data.resize(0);
				throw std::system_error(errno,
				    std::system_category(),
				    "Unable to read specified data");
			}
		} while ((total < data.size()) && (--receiveTries > 0));

		if (total < data.size())
			data.resize(total);
	}
}

ssize_t
Handle::write(
    const posixcpp::ByteArray & data,
    IOMode_t mode,
    time::Timespec timeout)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is null");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(descriptor() > 0, "file isn't open");

	// Write bytes count
	ssize_t count;

	if (_descriptor <= 0)
		throw std::runtime_error("File isn't opened");
	// Single write mode is set
	if (mode == IOMode_t::SINGLE) {
		count = ::write(_descriptor, data.data(), data.size());
		// Write was success
		if (count >= 0)
			return (count);
		else {
			throw std::system_error(errno, std::system_category(),
			    "Unable to write specified data");
			return (0);
		}
	} else {
		// Write tries remaining
		ssize_t sendTries = _writeTries;
		// Total write count
		size_t total = 0;
		do {
			Events src, res;
			bool evRes;
			do {
				res.clear();
				src.clear();
				src.setFlag(Events_t::OUTPUT);
				evRes = this->poll(src, res, timeout);
			} while (evRes && !res.testFlag(Events_t::OUTPUT));
			count = ::write(_descriptor, data.data() + total, data.size() - total);
			if (count > 0) {
				total += count;
				assert(total <= data.size());
			} else if ((errno != EAGAIN) && (errno != EWOULDBLOCK)
			    && (count != 0)) {
				throw std::system_error(errno,
				    std::system_category(),
				    "Unable to write specified data");
			}
		} while ((total < data.size()) && (sendTries-- > 0));
	}
	return (count);
}

ssize_t
Handle::write(const posixcpp::ByteArray &data)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is null");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(descriptor() > 0, "file isn't open");
	
	if (_descriptor <= 0)
		throw std::runtime_error("File isn't opened");
	
	return (::write(_descriptor, data.data(), data.size()));
}

ssize_t
Handle::write(const std::vector<Vector> &buffers)
{
	PRECONDITION_ASSERT(buffers.size() > 0, "data is empty");
	PRECONDITION_ASSERT(descriptor() > 0, "file isn't open");

	if (_descriptor <= 0)
		throw std::runtime_error("File isn't opened");
	
	return (::writev(this->_descriptor, reinterpret_cast<iovec*>(
	    const_cast<std::vector<Vector>&>(buffers).data()),
	    buffers.size()));
}

ssize_t
Handle::write(const std::vector<const posixcpp::ByteArray*> &buffers)
{
	PRECONDITION_ASSERT(buffers.size() > 0, "data is empty");
	PRECONDITION_ASSERT(descriptor() > 0, "file isn't open");

	if (_descriptor <= 0)
		throw std::runtime_error("File isn't opened");
	
	std::vector<Vector> v;
	v.reserve(buffers.size());
	for (const posixcpp::ByteArray *ba : buffers)
		v.emplace_back(*ba);
	
	return (::writev(this->_descriptor, reinterpret_cast<iovec*>(
	    const_cast<std::vector<Vector>&>(v).data()),
	    buffers.size()));
}

bool
Handle::poll(Events sevents, Events &revents, time::Timespec timeout)
{
	PRECONDITION_ASSERT(descriptor() > -1, "file isn't open");

	int tOut;
	int res;

	if (int64_t(timeout) >= 0)
		tOut = int(int64_t(timeout) / 1000000ll);
	else
		tOut = -1;

	pollfd thisFd = { .fd = _descriptor, .events = sevents.data() };
	res = ::poll(&thisFd, 1, tOut);
	if (res > 0) {
		revents.data() = thisFd.revents;
		return (true);
	} else if ((res == 0) || (errno == EINTR)) {
		return (false);
	} else {
		POSIX_CALL_ASSERT_X(res != -1, "Error while waiting for events on"
		    " the file");
		return (false);
	}
}

std::pair<Handle, Handle> Handle::pipe()
{
	std::pair<Handle, Handle> result;
	int fd[2];

	POSIX_CALL_ASSERT(::pipe(fd) != -1);
	
	result.first._descriptor = fd[0],
	result.first._descriptor = fd[1];

	return (result);
}

Pipe::Pipe()
{
	POSIX_CALL_ASSERT(::pipe(_pipeFds) == 0);
}

Pipe::Pipe(AccessMode_t aMode, BlockingMode_t bMode)
{
	POSIX_CALL_ASSERT(::pipe2(_pipeFds, int(aMode) | int(bMode)) == 0);
	_handle[0].setDescriptor(_pipeFds[0]);
	_handle[1].setDescriptor(_pipeFds[1]);
}

namespace posix
{
namespace io
{
void
pipe(Handle &first, Handle &second)
{
	int fd[2];

	POSIX_CALL_ASSERT(::pipe(fd) != -1);

	first._descriptor = fd[0];
	second._descriptor = fd[1];
}
}
}
