/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#include "posix_io_tty.hpp"

#include <unistd.h>
#include <sys/ioctl.h>

#include <unordered_map>

using namespace posix::io;
using namespace posix;

static std::unordered_map<TTY::BaudRate_t, std::string> br2str = {
	std::make_pair(TTY::BaudRate_t::BR_50, std::string("50")),
	std::make_pair(TTY::BaudRate_t::BR_75, std::string("75")),
	std::make_pair(TTY::BaudRate_t::BR_110, std::string("110")),
	std::make_pair(TTY::BaudRate_t::BR_134, std::string("134")),
	std::make_pair(TTY::BaudRate_t::BR_150, std::string("150")),
	std::make_pair(TTY::BaudRate_t::BR_300, std::string("300")),
	std::make_pair(TTY::BaudRate_t::BR_600, std::string("600")),
	std::make_pair(TTY::BaudRate_t::BR_1200, std::string("1200")),
	std::make_pair(TTY::BaudRate_t::BR_1800, std::string("1800")),
	std::make_pair(TTY::BaudRate_t::BR_2400, std::string("2400")),
	std::make_pair(TTY::BaudRate_t::BR_4800, std::string("4800")),
	std::make_pair(TTY::BaudRate_t::BR_9600, std::string("9600")),
	std::make_pair(TTY::BaudRate_t::BR_19200, std::string("19200")),
	std::make_pair(TTY::BaudRate_t::BR_38400, std::string("38400")),
	std::make_pair(TTY::BaudRate_t::BR_57600, std::string("57600")),
	std::make_pair(TTY::BaudRate_t::BR_115200, std::string("115200")),
	std::make_pair(TTY::BaudRate_t::BR_230400, std::string("230400"))
};

TTY::BaudRate_t
TTY::baudRateFromString(const std::string& str)
{	
	if (str == "134")
		return TTY::BaudRate_t::BR_134;
	else if (str == "150")
		return TTY::BaudRate_t::BR_150;
	else if (str == "300")
		return TTY::BaudRate_t::BR_300;
	else if (str == "600")
		return TTY::BaudRate_t::BR_600;
	else if (str == "1200")
		return TTY::BaudRate_t::BR_1200;
	else if (str == "1800")
		return TTY::BaudRate_t::BR_1800;
	else if (str == "4800")
		return TTY::BaudRate_t::BR_4800;
	else if (str == "9600")
		return TTY::BaudRate_t::BR_9600;
	else if (str == "19200")
		return TTY::BaudRate_t::BR_19200;
	else if (str == "38400")
		return TTY::BaudRate_t::BR_38400;
	else if (str == "57600")
		return TTY::BaudRate_t::BR_57600;
	else if (str == "115200")
		return TTY::BaudRate_t::BR_115200;
	else if (str == "230400")
		return TTY::BaudRate_t::BR_230400;
	else
		return TTY::BaudRate_t::BR_HANGUP;
}

TTY::BaudRate_t
TTY::baudRateFromInt(unsigned baud)
{
	switch (baud) {
	case 4800:
		return (TTY::BaudRate_t::BR_4800);
		break;
	case 9600:
		return (TTY::BaudRate_t::BR_9600);
		break;
	case 19200:
		return (TTY::BaudRate_t::BR_19200);
		break;
	case 38400:
		return (TTY::BaudRate_t::BR_38400);
		break;
	case 57600:
		return (TTY::BaudRate_t::BR_57600);
		break;
	default:
		return (TTY::BaudRate_t::BR_HANGUP);
	}
}

ssize_t
TTY::inputQueueSize()
{
	int result;

	POSIX_CALL_ASSERT(::ioctl(this->descriptor(), TIOCINQ, &result) != -1);

	return (result);
}

TTY::TTY() :
m_baudRate(BaudRate_t::BR_HANGUP)
{
	this->setCustomFlags(O_NOCTTY);
}

TTY::TTY(
    const std::string & fName,
    BaudRate_t bRate) :
File(fName),
m_baudRate(bRate)
{
	this->setCustomFlags(O_NOCTTY);
}

void
TTY::open()
{
	termios config;

	File::open();

	POSIX_CALL_ASSERT_X(::isatty(descriptor()) == 1, "File is not a tty");
	POSIX_CALL_ASSERT_X(::tcgetattr(descriptor(), &config) != -1,
	    "Unable to get termios attributes");

	config.c_cflag = CREAD | CLOCAL | CS8;

	config.c_iflag = ICRNL;
	config.c_oflag = 0;

	POSIX_CALL_ASSERT(::cfsetspeed(&config, uint(m_baudRate)) == 0);

	config.c_lflag = 0;
	cfmakeraw(&config);

	config.c_cc[VMIN] = 1;
	config.c_cc[VTIME] = 0;

	POSIX_CALL_ASSERT_X(::tcsetattr(descriptor(), TCSANOW, &config) != -1,
	    "Could not set termios attributes");
}

void
TTY::open(
    const std::string &fileName,
    Mode mode)
{
	this->setFileName(fileName);
	this->setMode(mode);

	this->open();
}

TTY::BaudRate_t
TTY::baudRate()
{
	return (m_baudRate);
}

void
TTY::setBaudRate(BaudRate_t newVal)
{	
	m_baudRate = newVal;
	if (descriptor() < 0)
		return;
	
	termios config;
	
	POSIX_CALL_ASSERT(::tcgetattr(descriptor(), &config) == 0);
	POSIX_CALL_ASSERT(::cfsetspeed(&config, uint(m_baudRate)) == 0);
}

void
TTY::flush(QueueSelector_t queues)
{
	POSIX_CALL_ASSERT(::tcflush(descriptor(), int(queues)) == 0);
}

void
TTY::drain()
{
	POSIX_CALL_ASSERT(::tcdrain(descriptor()) == 0);
}

void
TTY::flow(FlowControl_t fc)
{
	POSIX_CALL_ASSERT(::tcflow(descriptor(), int(fc)) == 0);
}

POSIX_BEGIN_NAMESPACE
	
namespace io
{

std::ostream & operator<<(std::ostream &stream, TTY::BaudRate_t value)
{
	switch (value) {
	case TTY::BaudRate_t::BR_150:
		stream << "150";
		break;
	case TTY::BaudRate_t::BR_300:
		stream << "300";
		break;
	case TTY::BaudRate_t::BR_600:
		stream << "600";
		break;
	case TTY::BaudRate_t::BR_1200:
		stream << "1200";
		break;
	case TTY::BaudRate_t::BR_4800:
		stream << "4800";
		break;
	case TTY::BaudRate_t::BR_9600:
		stream << "9600";
		break;
	case TTY::BaudRate_t::BR_19200:
		stream << "19200";
		break;
	case TTY::BaudRate_t::BR_38400:
		stream << "38400";
		break;
	case TTY::BaudRate_t::BR_57600:
		stream << "57600";
		break;
	case TTY::BaudRate_t::BR_115200:
		stream << "115200";
		break;
	case TTY::BaudRate_t::BR_HANGUP:
		stream << "Hangup";
		break;
	default:
		stream << "Undefined";
	}
	return (stream);
}

}

POSIX_END_NAMESPACE
