/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include "posix_signals.hpp"
#include "posixcpp.hpp"

#include <cstring>
#include <cassert>
#include <signal.h>
#include <unistd.h>

using posix::signals::Signal;

using namespace posix;
using namespace posix::signals;

Signal::Signal()
{
	std::memset(&m_sigInfo, 0, sizeof(m_sigInfo));
}

Signal::Signal(Number_t number)
{
	setSignumber(number);
}

Signal::Signal(const siginfo_t & siginfo) :
m_sigInfo(siginfo)
{
	assert(m_sigInfo.si_signo < SIGRTMAX);
}

Number_t
Signal::sigNumber() const
{
	return (Number_t(m_sigInfo.si_signo));
}

int
Signal::errNumber() const
{
	return (m_sigInfo.si_errno);
}

void *
Signal::pointer() const
{
	return (m_sigInfo.si_ptr);
}

void
Signal::setSignumber(Number_t value)
{
	m_sigInfo.si_signo = int(value);
}

Handler::Handler()
{
	_handlersInfo = new HandleInfo[SIGRTMAX];
}

Handler::~Handler()
{
	delete [] _handlersInfo;
}

void
Handler::switchSignalOn(Number_t signum, HandleType type)
{
	struct sigaction	sa;
	
	sa.sa_handler = NULL;
	sa.sa_sigaction = Handler::_action;
	sa.sa_flags = SA_SIGINFO;
	sigemptyset(&sa.sa_mask);
	sa.sa_restorer = NULL;
	
	_handlersInfo[size_t(signum)].handler = &Handler::signalAction;
	
	POSIX_CALL_ASSERT(::sigaction(int(signum), &sa, NULL) == 0);
}

void
Handler::sigQueue(Number_t signum)
{
	union sigval sv = { .sival_ptr = this };
	POSIX_CALL_ASSERT(::sigqueue(getpid(), int(signum), sv) == 0);
}

void
Handler::_action(
  int signo,
  siginfo_t *info,
  void *context)
{
	Signal		 sig(*info);
	Handler		*self = reinterpret_cast<Handler*>(sig.pointer());
	
	self->action(sig);
}

void
Handler::action(const Signal &sig)
{
	Number_t num = sig.sigNumber();
	
	if (_handlersInfo[size_t(num)].handler) {
		if (_handlersInfo[size_t(num)].type == BASIC)
			(this->*(_handlersInfo[size_t(num)].handler))(sig);
		else
			_signalsQueue.push_back(sig);
	}
}

namespace posix
{
namespace signals
{

std::ostream &
operator <<(
  std::ostream &stream,
  Number_t data)
{
	switch (data) {
	case Number_t::UNKNOWN:
		stream << "UNKNOWN"; break;
	case Number_t::HUP:
		stream << "HUP"; break;
	case Number_t::INT:
		stream << "INT"; break;
	case Number_t::QUIT:
		stream << "QUIT"; break;
	case Number_t::ILL:
		stream << "ILL"; break;
	case Number_t::KILL:
		stream << "KILL"; break;
	case Number_t::TRAP:
		stream << "TRAP"; break;
	case Number_t::ABRT:
		stream << "ABRT"; break;
	case Number_t::BUS:
		stream << "BUS"; break;
	case Number_t::FPE:
		stream << "FPE"; break;
	case Number_t::SEGV:
		stream << "SEGV"; break;
	case Number_t::USR1:
		stream << "USR1"; break;
	case Number_t::USR2:
		stream << "USR2"; break;
	case Number_t::TERM:
		stream << "TERM"; break;
	default:
		stream << "UNDEFINED"; break;
	}
	stream << " (" << strsignal(int(data)) << ")";

	return (stream);
}

std::ostream &
operator <<(
  std::ostream &stream,
  const Signal &entity)
{
	stream << "Signal:\n\tNumber: " << Number_t(entity.m_sigInfo.si_signo) <<
	    "\tErrno:  " << strerror(entity.m_sigInfo.si_errno) << '\n';

	return (stream);
}

}
}
