/*
    This file is part of Posixcpp library.

    Posixcpp library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    Posixcpp library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with Posixcpp library. If not, see
    <http://www.gnu.org/licenses/>.
 */

#include "posix_sockets_socket.hpp"

#include <cstring>
#include <list>
#include <ifaddrs.h>

using namespace posix::io::sockets;
using namespace posix::io;
using namespace posix;

Address::Address(sockaddr* addr, Family_t family)
{
	addr->sa_family = int(family);
}

Message::Message()
{
	std::memset(&_data, 0, sizeof (_data));
}

void
Message::setAdress(const Address &addr)
{
	_address.resize(addr.length());
	_address.copyExistingData((uint8_t*)const_cast<Address&>(addr).handle(),
	    addr.length());
	
	_data.msg_name = _address.data();
	_data.msg_namelen = addr.length();
}

void
Message::addBuffer(const Vector &buf)
{
	_buffers.emplace_back(buf);
	// It's important to place the buffer pointer from vector
	// after each insertion as it may be reallocated
	_data.msg_iov = reinterpret_cast<iovec*>(_buffers.data());
	_data.msg_iovlen = _buffers.size();
}

size_t
Message::size() const
{
	size_t result = 0;
	
	for (const Vector &v : _buffers)
		result += v.data().iov_len;
	
	return (result);
}

PhisicalAddress::PhisicalAddress(uint16_t ethernetProtocol) :
Address((sockaddr*) (&m_address), Family_t::PACKET)
{
	this->init(ethernetProtocol);
}

void
PhisicalAddress::init(uint16_t ethernetProtocol)
{
	this->m_address.sll_ifindex = 0;
	this->m_address.sll_protocol = htobe16(ethernetProtocol);
	this->m_address.sll_halen = ETH_ALEN;
	this->m_address.sll_pkttype = PACKET_BROADCAST;
	memset(this->m_address.sll_addr, 0xFF, ETH_ALEN);
}

socklen_t
PhisicalAddress::length() const
{
	return sizeof (struct sockaddr_ll);
}

InetAddress::InetAddress(const sockaddr_in &addr) :
Address((sockaddr*) (&m_address), Family_t::INET),
m_address(addr)
{
}

InetAddress::InetAddress(in_addr_t addr, in_port_t port) :
Address((sockaddr*) (&m_address), Family_t::INET)
{
	this->init(addr, port);
}

InetAddress
InetAddress::networkAddress() const
{
	InetAddress result(::inet_netof(m_address.sin_addr));
	
	return (result);
}

unsigned char* PhisicalAddress::address()
{
	return m_address.sll_addr;
}

InetAddress::InetAddress(const std::string& addr) :
Address((sockaddr*) (&m_address), Family_t::INET)
{
	setAddress(addr);
}

bool
InetAddress::operator==(const InetAddress &rhs) const
{
	return ((m_address.sin_family == rhs.m_address.sin_family) &&
		(m_address.sin_addr.s_addr == rhs.m_address.sin_addr.s_addr) &&
		(m_address.sin_port == rhs.m_address.sin_port));
}

void
InetAddress::setAddress(const std::string& addr)
{
	size_t semicolonPos = addr.find(':');

	in_port_t port;
	std::string str = addr.substr(0, semicolonPos);
	POSIX_CALL_ASSERT_X(::inet_aton(str.c_str(), &m_address.sin_addr) != 0,
	    std::string("Invalid address ") + str);
	if (semicolonPos != std::string::npos) {
		str = addr.substr(semicolonPos + 1);
		port = std::atoi(str.c_str());
		this->setPort(port);
	}
}

void
InetAddress::init(in_addr_t addr, in_port_t port)
{
	this->setPort(port);
	this->setIp(addr);
}

socklen_t
InetAddress::length() const
{
	return sizeof (sockaddr_in);
}

InetAddress::List
InetAddress::addressList()
{
	struct ifaddrs *ifAddrStruct;
	List result;
	
	::getifaddrs(&ifAddrStruct);
	// перебираем все сетевые инт. host'a
	for (struct ifaddrs *ifa = ifAddrStruct; ifa != NULL;
	    ifa= ifa->ifa_next) {
		if (!ifa->ifa_addr)
			continue;
		// если IPv4
		if (ifa->ifa_addr->sa_family == AF_INET) {
			sockaddr_in *inetAddr = reinterpret_cast<sockaddr_in*> (
			    ifa->ifa_addr);
			result.emplace_back(*inetAddr);
		}
	}
	
	return (result);
}

Socket::Socket(
    Family_t domain,
    Type_t type,
    int protocol)
{
	this->m_domain = domain;
	this->m_type = type;
	this->m_protocol = htobe16(protocol);
}

void
Socket::create()
{
	int desc;
	
	POSIX_CALL_ASSERT((desc = ::socket(int(this->m_domain),
	    int(this->m_type), m_protocol)) >= 0);
	
	setDescriptor(desc);
}

void
Socket::bind(const Address& addr)
{
	PRECONDITION_ASSERT(descriptor() > 0, "socket isn't open");
	
	POSIX_CALL_ASSERT(::bind(descriptor(), const_cast<Address&> (addr).handle(),
	    addr.length()) == 0);
}

size_t
Socket::send(
    const posixcpp::ByteArray& data,
    Flags_t flags)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is empty");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(descriptor() > 0, "socket isn't open");

	// Количество записанных байт
	ssize_t count;

	POSIX_CALL_ASSERT((count = ::send(descriptor(), data.data(),
	    data.size(), int(flags))) != -1);

	POSTCONDITION_ASSERT(count <= ssize_t(data.size()), "Invalid result");
	if (count < 0)
		count = 0;

	return (size_t(count));
}

size_t
Socket::send(
    const Message &data,
    Flags_t flags)
{
	PRECONDITION_ASSERT(descriptor() > 0, "socket isn't open");

	// Written data size
	ssize_t count;

	POSIX_CALL_ASSERT((count = ::sendmsg(descriptor(), &data.data(),
	    int(flags))) != -1);

	POSTCONDITION_ASSERT(count <= ssize_t(data.size()), "Invalid result");
	if (count < 0)
		count = 0;

	return (size_t(count));
}

size_t
Socket::sendTo(
    const posixcpp::ByteArray& data,
    const Address& addr,
    Flags_t flags)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is empty");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(descriptor() > 0, "socket isn't open");

	// Количество записанных байт
	ssize_t count;

	POSIX_CALL_ASSERT((count = ::sendto(descriptor(), data.data(),
	    data.size(), int(flags), const_cast<Address&> (addr).handle(),
	    addr.length())) != -1);

	POSTCONDITION_ASSERT(count <= ssize_t(data.size()), "Invalid result");
	if (count < 0)
		count = 0;

	return count;
}

void
Socket::receive(
    posixcpp::ByteArray& data,
    Flags_t flags)
{
	PRECONDITION_ASSERT(!data.isNull(), "data is empty");
	PRECONDITION_ASSERT(data.size() > 0, "data is empty");
	PRECONDITION_ASSERT(descriptor() > 0, "socket isn't open");

	ssize_t count;

	POSIX_CALL_ASSERT((count = ::recv(descriptor(), data.data(),
	    data.size(), int(flags))) != -1);

	POSTCONDITION_ASSERT(count <= ssize_t(data.size()), "Invalid result");
	if (count < 0)
		count = 0;

	data.resize(count);
}

void
Socket::listen()
{
	POSIX_CALL_ASSERT(::listen(descriptor(), 0) != -1);
}

void
Socket::accept(Socket &sock)
{
	int desc;

	POSIX_CALL_ASSERT((desc = ::accept(descriptor(), NULL, 0)) > 0);

	sock.setBlockingMode(this->blockingMode());
	sock.setProtocol(this->protocol());
	sock.setReadTries(this->readTries());
	sock.settype(this->type());
	sock.setDescriptor(desc);
}

Socket
Socket::accept()
{
	Socket	res(this->m_domain, this->m_type, this->m_protocol);
	this->accept(res);
	return (res);
}

void
Socket::shutdown(ShutdownType_t how)
{
	POSIX_CALL_ASSERT((::shutdown(descriptor(), int(how))) != -1);
}

void
Socket::connect(const Address &addr)
{
	POSIX_CALL_ASSERT((::connect(descriptor(),
	    const_cast<Address&> (addr).handle(), addr.length())) != -1);
}

namespace posix
{
namespace io
{
namespace sockets
{

std::ostream & operator <<(std::ostream &stream, const io::sockets::InetAddress
    &addr)
{
	std::string str(inet_ntoa(addr.address().sin_addr));
	stream << str << ':' << addr.port();

	return (stream);
}

std::pair<Socket, Socket>
pair(Family_t domain, Type_t type, int protocol)
{
	std::pair<Socket, Socket> result = { Socket(domain, type),
	    Socket(domain, type) };
	
	int fds[2];
	
	POSIX_CALL_ASSERT(::socketpair(int(domain), int(type),
	    htobe16(protocol), fds) != -1);
	result.first.setDescriptor(fds[0]);
	result.second.setDescriptor(fds[1]);
	
	return (result);
}

}
}
}
