/*
	This file is part of Posixcpp library.

	Posixcpp library is free software: you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public License
	as published by the Free Software Foundation, either version 3 of
	the License, or (at your option) any later version.

	Posixcpp library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with Posixcpp library. If not, see
	<http://www.gnu.org/licenses/>.
 */

#include "posix_time_clock.hpp"

#include <iostream>

using namespace posix::time;

std::map<Clock_t, Clock> Clock::m_clocks;
Clock::StaticInitializer Clock::m_initializer;

Clock::StaticInitializer::StaticInitializer( )
{
	Clock::m_clocks.emplace(Clock_t::MONOTONIC, Clock_t::MONOTONIC);
	Clock::m_clocks.emplace(Clock_t::MONOTONIC_RAW, Clock_t::MONOTONIC_RAW);
	Clock::m_clocks.emplace(Clock_t::MONOTONIC_COARSE, Clock_t::MONOTONIC_COARSE);
	Clock::m_clocks.emplace(Clock_t::PROCESS_CPU, Clock_t::PROCESS_CPU);
	Clock::m_clocks.emplace(Clock_t::REALTIME, Clock_t::REALTIME);
	Clock::m_clocks.emplace(Clock_t::REALTIME_COARSE, Clock_t::REALTIME_COARSE);
	Clock::m_clocks.emplace(Clock_t::THREAD_CPU, Clock_t::THREAD_CPU);

}

Clock::StaticInitializer::~StaticInitializer( )
{
}

Clock::Clock( Clock_t clockId ) :
m_clockId( clockId )
{

}

Clock::~Clock( )
{

}

Timespec Clock::getRes( ) const
{
	Timespec res;

	POSIX_CALL_ASSERT_X(clock_getres(clockid_t(m_clockId), &res.data()) == 0,
					"Unable to get clock resolution");

	return res;
}

Timespec Clock::getTime( ) const
{
	Timespec res;
	
	POSIX_CALL_ASSERT_X(clock_gettime(clockid_t(m_clockId), &res.data()) == 0,
					"Unable to get time");
	
	return res;
}

void Clock::setTime( const Timespec& time )
{
	POSIX_CALL_ASSERT_X(clock_settime(clockid_t(m_clockId), &time.data()) == 0,
					"Unable to set time");
}

Timespec
Clock::nanoSleep(
  const Timespec &required,
  SleepMode_t mode) const
{
	int		res;
	Timespec	result;
	
	if (mode == SleepMode_t::RELATIVE) {
		Timespec	remain;
		res = clock_nanosleep(clockid_t(m_clockId), int(mode), &required.data(),
		    &remain.data());
		result = remain;
	} else {
		res = clock_nanosleep(clockid_t(m_clockId), int(mode), &required.data(),
		    NULL);
		result = required;
	}
	
	POSIX_CALL_ASSERT_X((res == 0) || (res == EINTR),
	    "Unable to complete clock_nanosleep");
	
	return(result);
}

namespace posix
{
namespace time
{

Clock&
clock(Clock_t id)
{
	std::map<Clock_t, Clock>::iterator it = Clock::m_clocks.find(id);
	if (it != Clock::m_clocks.end())
		return it->second;
	else
		throw std::runtime_error("No such clock");
}

Timespec
nanoSleep(const Timespec &required)
{
	Timespec	remain;
	int		result;
	
	result = ::nanosleep(&required.data(), &remain.data());
	if (result == -1 && errno != EINTR)
		throw std::system_error(errno, std::system_category(),
		    __PRETTY_FUNCTION__);

	return (remain);
}

}

}
